#lang racket

(require compatibility/mlist)

(struct ap (π vs ρ ρs σ κs τ ρ0))        ; application pattern
(struct ev (e ρ ρs σ κs τ ρ0))           ; evaluation pattern
(struct ko (κ v ρ ρs σ κs))              ; continuation pattern

(struct defk (id))                       ; define form continuation
(struct evalk (τ ρ0))                    ; evaluate form continuation
(struct fchk (π vs ρ0))                  ; for-each native continuation
(struct ifk (e τ ρ0))                    ; if form continuation
(struct ifek (e1 e2 τ ρ0))               ; if-else form continuation
(struct mppk (π vs vs* ρ0))              ; map native  continuation
(struct multiplek (v vs τ ρ0))           ; multiple argument continuation
(struct nullk (τ ρ0))                    ; no argument continuation
(struct randk (π rands vs τ ρ0))         ; operand continuation
(struct ratork (rands τ ρ0))             ; operator continuation
(struct repk ())                         ; REP continuation
(struct retk (ρ ρs))                     ; return continuation
(struct seqk (es τ ρ0))                  ; sequence continuation
(struct setk (id ρ0))                    ; set form continuation
(struct singlek (v τ ρ0))                ; single argument continuation

(struct apl ())                          ; apply native
(struct ccc ())                          ; call-with-current-continuation native
(struct clo (ps es ρs))                  ; closure
(struct con (ρ ρs κs))                   ; continuation
(struct fch ())                          ; for-each native
(struct mpp ())                          ; map native
(struct nat (π))                         ; regular native
(struct nyi ())                          ; not yet implemented native
(struct rea ())                          ; read native
(struct undefined ())                    ; undefined pattern
(struct unspecified ())                  ; unspecified pattern

; literal cloning

(define (lit e)
  (match e
    ((and e (? string?))
     (string-copy e))
    ((cons car cdr)
     (mcons (lit car) (lit cdr)))    
    ((vector es ...)
     (list->vector (map lit es)))
    (_
     e)))

; error message with symbol

(define (msg ms id)
  (string-append ms (symbol->string id)))

; escape to read-eval-print

(define (rep ms ρ σ)
  (ko (repk) ms ρ '() σ '()))

; expression sequence

(define (seq es ρ ρs σ κs τ ρ0)
  (match es
    ((list e) 
     (ev e ρ ρs σ κs τ ρ0))
    ((cons e es) 
     (ev e ρ ρs σ (cons (seqk es τ ρ0) κs) #f ρ0))
    (_
     (rep "expression sequence required" σ ρ0))))

; thunk

(define (thk e ρ ρs σ κs τ ρ0)
  (if τ 
      (ev e '() (cons ρ ρs) σ κs #t ρ0)
      (ev e '() (cons ρ ρs) σ (cons (retk ρ ρs) κs) #t ρ0)))

; step function

(define (step state)
  (match state 
    
    ; variable reference
    
    ((ev (and e (? symbol?)) ρ ρs σ (cons κ κs) τ ρ0)
     (let loop ((ρ* ρ) 
                (ρs* ρs))
       (match (assoc e ρ*)
         ((cons _ a) 
          (ko κ (cdr (assoc a σ)) ρ ρs σ κs))
         (_ 
          (match ρs*
            ((cons ρ ρs)
             (loop ρ ρs))
            (_
             (ko κ (undefined) ρ ρs σ κs)))))))         
    
    ; expression sequence
    
    ((ev `(begin ,es ...) ρ ρs σ κs τ ρ0)
     (seq es ρ ρs σ κs τ ρ0))
    
    ((ko (seqk es τ ρ0) _ ρ ρs σ κs)
     (seq es ρ ρs σ κs τ ρ0))
    
    ; definition
    
    ((ev `(define (,id . ,ps) ,es ...) ρ ρs σ (cons κ κs) _ ρ0)
     (match (assoc id ρ)
       ((cons _ a)
        (let* ((clo (clo ps es (cons ρ ρs)))
               (σ (cons (cons a clo) σ)))
          (ko κ clo ρ ρs σ κs)))
       (_ 
        (let* ((fresh (gensym))
               (ρ (cons (cons id fresh) ρ))
               (clo (clo ps es (cons ρ ρs)))
               (σ (cons (cons fresh clo) σ)))
          (ko κ clo ρ ρs σ κs)))))
    
    ; ---
    
    ((ev `(define ,id ,e) ρ ρs σ κs _ ρ0)
     (ev e ρ ρs σ (cons (defk id) κs) #f ρ0))
    
    ((ko (defk id) v ρ ρs σ (cons κ κs))
     (match (assoc id ρ)
       ((cons _ a)
        (let ((σ* (cons (cons a v) σ)))
          (ko κ v ρ ρs σ* κs)))
       (_ 
        (let* ((fresh (gensym))
               (ρ* (cons (cons id fresh) ρ))
               (σ* (cons (cons fresh v) σ)))
          (ko κ v ρ* ρs σ* κs)))))
    
    ; evaluation
    
    ((ev `(evaluate ,e) ρ ρs σ κs τ ρ0)
     (ev e ρ ρs σ (cons (evalk τ ρ0) κs) τ ρ0))
    
    ((ko (evalk τ ρ0) v ρ ρs σ κs)
     (ev v ρ ρs σ κs τ ρ0))
    
    ; conditional
    
    ((ev `(if ,e ,e1) ρ ρs σ κs τ ρ0)
     (ev e ρ ρs σ (cons (ifk e1 τ ρ0) κs) #f ρ0))
    
    ((ko (ifk _ _ _) #f ρ ρs σ (cons κ κs))
     (ko κ (unspecified) ρ ρs σ κs))
    
    ((ko (ifk e1 τ ρ0) _ ρ ρs σ κs)
     (thk e1 ρ ρs σ κs τ ρ0))
    
    ; ---
    
    ((ev `(if ,e ,e1 ,e2) ρ ρs σ κs τ ρ0)
     (ev e ρ ρs σ (cons (ifek e1 e2 τ ρ0) κs) #f ρ0))
    
    ((ko (ifek _ e2 τ ρ0) #f ρ ρs σ κs)
     (thk e2 ρ ρs σ κs τ ρ0))
    
    ((ko (ifek e1 _ τ ρ0) _ ρ ρs σ κs)
     (thk e1 ρ ρs σ κs τ ρ0))
    
    ; abstraction
    
    ((ev `(lambda ,ps ,es ...) ρ ρs σ (cons κ κs) _ _)
     (let ((clo* (clo ps es (cons ρ ρs))))
       (ko κ clo* ρ ρs σ κs)))
    
    ; load
    
    ((ev `(load e) _ σ _ _ _ ρ0)
     (rep "not yet implemented: load" σ ρ0))
    
    ; quotation
    
    ((ev `(quote ,e) ρ ρs σ (cons κ κs) _ _)
     (ko κ (lit e) ρ ρs σ κs))
    
    ; assignment
    
    ((ev `(set! ,id ,e) ρ ρs σ κs _ ρ0)
     (ev e ρ ρs σ (cons (setk id ρ0) κs) #f ρ0))
    
    ((ko (setk id ρ0) v ρ ρs σ (cons κ κs))
     (let loop ((ρ* ρ) 
                (ρs* ρs))
       (match (assoc id ρ*)
         ((cons _ a) 
          (ko κ v ρ ρs (cons (cons a v) σ) κs))
         (_ 
          (match ρs*
            ((cons ρ ρs)
             (loop ρ ρs))
            (_
             (rep (msg "variable not found: " id) σ ρ0)))))))
    
    ; application
    
    ((ev `(,rator) ρ ρs σ κs τ ρ0)
     (ev rator ρ ρs σ (cons (nullk τ ρ0) κs) τ ρ0))
    
    ((ko (nullk τ ρ0) π ρ ρs σ κs)
     (ap π '() ρ ρs σ κs τ ρ0))
    
    ; ---
    
    ((ev `(,rator . ,rands) ρ ρs σ κs τ ρ0)
     (ev rator ρ ρs σ (cons (ratork rands τ ρ0) κs) #f ρ0))
    
    ((ko (ratork (list rand) τ ρ0) π ρ ρs σ κs)
     (ev rand ρ ρs σ (cons (singlek π τ ρ0) κs) τ ρ0))    
    
    ((ko (singlek π τ ρ0) v ρ ρs σ κs)
     (ap π (list v) ρ ρs σ κs τ ρ0))
    
    ((ko (ratork (cons rand rands) τ ρ0) π ρ ρs σ κs)
     (ev rand ρ ρs σ (cons (randk π rands '() τ ρ0) κs) #f ρ0))    
    
    ((ko (randk π (list rand) vs τ ρ0) v ρ ρs σ κs)
     (ev rand ρ ρs σ (cons (multiplek π (cons v vs) τ ρ0) κs) τ ρ0))   
    
    ((ko (randk π (cons rand rands) vs τ ρ0) v ρ ρs σ κs)
     (ev rand ρ ρs σ (cons (randk π rands (cons v vs) τ ρ0) κs) #f ρ0))   
    
    ((ko (multiplek π vs τ ρ0) v ρ ρs σ κs)
     (ap π (reverse (cons v vs)) ρ ρs σ κs τ ρ0))
    
    ; ---
    
    ((ap (clo ps es ρs#) vs ρ ρs σ κs τ ρ0)
     (let ((κs* (if τ κs (cons (retk ρ ρs) κs))))
       (let loop ((ps* ps) 
                  (vs* vs) 
                  (ρ# '()) 
                  (σ* σ))
         (match (cons ps* vs*)
           ((cons '() '()) 
            (seq es ρ# ρs# σ* κs* #t ρ0))
           ((cons (cons p ps*) (cons v vs*))
            (let* ((fresh (gensym))
                   (ρ# (cons (cons p fresh) ρ#))
                   (σ* (cons (cons fresh v) σ*)))   
              (loop ps* vs* ρ# σ*)))
           ((cons (and ps* (? symbol?)) vs*)
            (let* ((fresh (gensym))
                   (ρ# (cons (cons ps* fresh) ρ#))
                   (σ* (cons (cons fresh (list->mlist vs*)) σ*)))   
              (seq es ρ# ρs# σ* κs* #t ρ0)))
           (_
            (rep "non-matching argument list" σ* ρ0))))))
    
    ((ap (con ρ ρs (cons κ κs)) (mlist v) _ _ σ _ _ ρ0)
     (ko κ v ρ ρs σ κs))
    
    ((ap (con _ _ _) _ _ _ σ _ _ ρ0)
     (rep "invalid arguments for: call/cc" σ ρ0))
    
    ((ap (nat π) vs ρ ρs σ (cons k κs) _ _)
     (ko k (apply π vs) ρ ρs σ κs))
    
    ((ko (retk ρ ρs) v _ _ σ (cons κ κs))
     (ko κ v ρ ρs σ κs))
    
    ; natives
    
    ((ap (apl) (mlist π vs) ρ ρs σ κs τ ρ0)
     (ap π vs ρ ρs σ κs τ ρ0))
    
    ((ap (ccc) (mlist π) ρ ρs σ κs τ ρ0)
     (ap π (mlist (con ρ ρs κs)) ρ ρs σ κs τ ρ0))
    
    ((ap (fch) (mlist π '()) ρ ρs σ κs _ ρ0)
     (ko (car κs) '() ρ ρs σ (cdr κs)))
    
    ((ap (fch) (mlist π (mcons v vs)) ρ ρs σ κs _ ρ0)
     (ap π (mlist v) ρ ρs σ (cons (fchk π vs ρ0) κs) #f ρ0))
    
    ((ko (fchk π '() ρ0) _ ρ ρs σ (cons κ κs))   
     (ko κ (unspecified) ρ ρs σ κs))
    
    ((ko (fchk π (mcons v vs) ρ0) _ ρ ρs σ κs)   
     (ap π (mlist v) ρ ρs σ (cons (fchk π vs ρ0) κs) #f ρ0))
    
    ((ap (mpp) (mlist π '()) ρ ρs σ κs _ ρ0)
     (ko (car κs) '() ρ ρs σ (cdr κs)))
    
    ((ap (mpp) (mlist π (mcons v vs)) ρ ρs σ κs _ ρ0)
     (ap π (mlist v) ρ ρs σ (cons (mppk π vs '() ρ0) κs) #f ρ0))
    
    ((ko (mppk π '() vs* _) v* ρ ρs σ (cons κ κs))   
     (ko κ (mreverse (mcons v* vs*)) ρ ρs σ κs))
    
    ((ko (mppk π (mcons v vs) vs* ρ0) v* ρ ρs σ κs)   
     (ap π (mlist v) ρ ρs σ (cons (mppk π vs (mcons v* vs*) ρ0) κs) #f ρ0))
    
    ((ap (nyi) _ _ _ σ _ _ ρ0)
     (rep "not yet implemented" σ ρ0))    
    
    ((ap (rea) '() ρ ρs σ κs _ ρ0)
     (ko (car κs) (lit (read)) ρ ρs σ κs))
    
    ((ap _ _ _ _ σ _ _ ρ0)
     (rep "invalid application" σ ρ0))        
    
    ; literal
    
    ((ev e ρ ρs σ (cons κ κs) _ _)
     (ko κ (lit e) ρ ρs σ κs)) 
    
    ; read-eval-print
    
    ((ko (repk) v ρ ρs σ _)
     (display v) 
     (newline)
     (display "> ") 
     (ev (read) ρ ρs σ (list (repk)) #f ρ))))

; initialization

(define native-names '(-
                       *
                       /
                       +
                       <
                       <=
                       =
                       >
                       >=
                       abs
                       acos
                       ;angle
                       append
                       apply
                       asin
                       assoc
                       assq
                       ;assv
                       atan
                       boolean?
                       caaaar
                       caaadr
                       caaar
                       caadar
                       caaddr
                       caadr
                       caar
                       cadaar
                       cadadr
                       cadar
                       caddar
                       cadddr
                       caddr
                       cadr
                       call-with-current-continuation
                       ;call-with-input-file
                       ;call-with-output-file
                       car
                       cdaaar
                       cdaadr
                       cdaar
                       cdadar
                       cdaddr
                       cdadr
                       cdar
                       cddaar
                       cddadr
                       cddar
                       cdddar
                       cddddr
                       cdddr
                       cddr
                       cdr
                       ;ceiling
                       char->integer
                       ;char-alphabetic?
                       ;char-ci<?
                       ;char-ci<=?
                       ;char-ci=?
                       ;char-ci>?
                       ;char-ci>=?
                       ;char-downcase
                       ;char-lower-case?
                       ;char-numeric?
                       ;char-ready?
                       ;char-upcase
                       ;char-upper-case?
                       ;char-whitespace?
                       char?
                       char<?
                       char<=?
                       char=?
                       char>?
                       char>=?
                       close-input-port
                       close-output-port
                       ;complex?
                       cons
                       cos
                       ;current-input-port
                       ;current-output-port
                       ;denominator
                       display
                       eof-object?
                       eq?
                       equal?
                       ;eqv?
                       ;eval
                       even?
                       ;exact->inexact
                       ;exact?
                       exp
                       expt
                       ;floor
                       for-each
                       force
                       ;gcd
                       ;imag-part
                       ;inexact->exact
                       ;inexact?
                       input-port?
                       ;integer->char
                       integer?
                       length
                       list
                       ;list->string
                       list->vector
                       ;list-ref
                       ;list-tail
                       ;load
                       log
                       ;magnitude
                       ;make-polar
                       ;make-rectangular
                       make-string
                       make-vector
                       map
                       max
                       member
                       memq
                       ;memv
                       min
                       modulo
                       negative?
                       newline
                       not
                       null?
                       number->string
                       number?
                       ;numerator
                       odd?
                       open-input-file
                       open-output-file
                       output-port?
                       pair?
                       ;THREAD_PEEK-char
                       positive?
                       procedure?
                       quotient
                       ;rational?
                       ;rationalize
                       read
                       read-char
                       ;real-part
                       real?
                       remainder
                       reverse
                       round
                       ;scheme-report-environment
                       set-car!
                       set-cdr!
                       sin
                       sqrt
                       ;string
                       ;string->list
                       string->number
                       string->symbol
                       string-append
                       ;string-ci<?
                       ;string-ci<=?
                       ;string-ci=?
                       ;string-ci>?
                       ;string-ci>=?
                       string-copy
                       ;string-fill!
                       string-length
                       string-ref
                       string-set!
                       string?
                       string<?
                       string<=?
                       string=?
                       string>?
                       string>=?
                       substring
                       symbol->string
                       symbol?
                       tan
                       ;transcript-off
                       ;transcript-on
                       truncate
                       vector
                       vector->list
                       ;vector-fill!
                       vector-length
                       vector-ref
                       vector-set!
                       vector?
                       ;with-input-from-file
                       ;with-output-to-file
                       write
                       write-char
                       zero?
                       ; STREAMS
                       stream-null?
                       the-empty-stream
                       cons-stream
                       stream-car
                       stream-cdr
                       ; BYTEVECTORS
                       bytevector?
                       make-bytevector
                       bytevector
                       bytevector-length
                       bytevector-ref
                       bytevector-set!
                       list->bytevector
                       ; EXTENSIONS
                       circularity-level
                       clock
                       collect
                       error
                       exit
                       false
                       format
                       gensym
                       random
                       true))

(define native-implementations (list (nat -)
                                     (nat *)
                                     (nat /)
                                     (nat +)
                                     (nat <)
                                     (nat <=)
                                     (nat =)
                                     (nat >)
                                     (nat >=)
                                     (nat abs)
                                     (nat acos)
                                     ;(nat angle)
                                     (nat append)
                                     (apl)	
                                     (nat asin)
                                     (nat massoc)
                                     (nat massq)
                                     ;(nat assv)
                                     (nat atan)
                                     (nat boolean?)
                                     (nyi)	
                                     (nyi)	
                                     (nyi)	
                                     (nyi)	
                                     (nyi)	
                                     (nyi)	
                                     (nyi)	
                                     (nyi)	
                                     (nyi)	
                                     (nyi)	
                                     (nyi)	
                                     (nyi)	
                                     (nyi)	
                                     (nyi)	
                                     (ccc)	
                                     ;(nat call-with-input-file)
                                     ;(nat call-with-output-file)
                                     (nat mcar)
                                     (nyi)	
                                     (nyi)	
                                     (nyi)	
                                     (nyi)	
                                     (nyi)	
                                     (nyi)	
                                     (nyi)	
                                     (nyi)	
                                     (nyi)	
                                     (nyi)	
                                     (nyi)	
                                     (nyi)	
                                     (nyi)	
                                     (nyi)	
                                     (nat mcdr)
                                     ;(nat ceiling)
                                     (nat char->integer)
                                     ;(nat char-alphabetic?)
                                     ;(nat char-ci<?)
                                     ;(nat char-ci<=?)
                                     ;(nat char-ci=?)
                                     ;(nat char-ci>?)
                                     ;(nat char-ci>=?)
                                     ;(nat char-downcase)
                                     ;(nat char-lower-case?)
                                     ;(nat char-numeric?)
                                     ;(nat char-ready?)
                                     ;(nat char-upcase)
                                     ;(nat char-upper-case?)
                                     ;(nat char-whitespace?)
                                     (nat char?)
                                     (nat char<?)
                                     (nat char<=?)
                                     (nat char=?)
                                     (nat char>?)
                                     (nat char>=?)
                                     (nat close-input-port)
                                     (nat close-output-port)
                                     ;(nat complex?)
                                     (nat mcons)
                                     (nat cos)
                                     ;(nat current-input-port)
                                     ;(nat current-output-port)
                                     ;(nat denominator)
                                     (nat display)
                                     (nat eof-object?)
                                     (nat eq?)
                                     (nat equal?)
                                     ;(nat eqv?)
                                     ;(nat eval)
                                     (nat even?)
                                     ;(nat exact->inexact)
                                     ;(nat exact?)
                                     (nat exp)
                                     (nat expt)
                                     ;(nat floor)
                                     (fch)	
                                     (nat force)
                                     ;(nat gcd)
                                     ;(nat imag-part)
                                     ;(nat inexact->exact)
                                     ;(nat inexact?)
                                     (nat input-port?)
                                     ;(nat integer->char)
                                     (nat integer?)
                                     (nat length)
                                     (nat mlist)
                                     ;(nat list->string)
                                     (nat list->vector)
                                     ;(nat list-ref)
                                     ;(nat list-tail)
                                     ;(nat load)
                                     (nat log)
                                     ;(nat magnitude)
                                     ;(nat make-polar)
                                     ;(nat make-rectangular)
                                     (nat make-string)
                                     (nat make-vector)
                                     (mpp)	
                                     (nat max)
                                     (nat mmember)
                                     (nat mmemq)
                                     ;(nat memv)
                                     (nat min)
                                     (nat modulo)
                                     (nat negative?)
                                     (nat newline)
                                     (nat not)
                                     (nat null?)
                                     (nat number->string)
                                     (nat number?)
                                     ;(nat numerator)
                                     (nat odd?)
                                     (nat open-input-file)
                                     (nat open-output-file)
                                     (nat output-port?)
                                     (nat mpair?)
                                     ;(nat peek-char)
                                     (nat positive?)
                                     (nat procedure?)
                                     (nat quotient)
                                     ;(nat rational?)
                                     ;(nat rationalize)
                                     (rea)	
                                     (nat read-char)
                                     ;(nat real-part)
                                     (nat real?)
                                     (nat remainder)
                                     (nat reverse)
                                     (nat round)
                                     ;(nat scheme-report-environment)
                                     (nat set-mcar!)
                                     (nat set-mcdr!)
                                     (nat sin)
                                     (nat sqrt)
                                     ;(nat string)
                                     ;(nat string->list)
                                     (nat string->number)
                                     (nat string->symbol)
                                     (nat string-append)
                                     ;(nat string-ci<?)
                                     ;(nat string-ci<=?)
                                     ;(nat string-ci=?)
                                     ;(nat string-ci>?)
                                     ;(nat string-ci>=?)
                                     (nat string-copy)
                                     ;(nat string-fill!)
                                     (nat string-length)
                                     (nat string-ref)
                                     (nat string-set!)
                                     (nat string?)
                                     (nat string<?)
                                     (nat string<=?)
                                     (nat string=?)
                                     (nat string>?)
                                     (nat string>=?)
                                     (nat substring)
                                     (nat symbol->string)
                                     (nat symbol?)
                                     (nat tan)
                                     ;(nat transcript-off)
                                     ;(nat transcript-on)
                                     (nat truncate)
                                     (nat vector)
                                     (nat vector->list)
                                     ;(nat vector-fill!)
                                     (nat vector-length)
                                     (nat vector-ref)
                                     (nat vector-set!)
                                     (nat vector?)
                                     ;(nat with-input-from-file)
                                     ;(nat with-output-to-file)
                                     (nat write)
                                     (nat write-char)
                                     (nat zero?)
                                     ;		
                                     (nyi)	
                                     (nyi)	
                                     (nyi)	
                                     (nyi)	
                                     (nyi)	
                                     ;		
                                     (nyi)	
                                     (nyi)	
                                     (nyi)	
                                     (nyi)	
                                     (nyi)	
                                     (nyi)	
                                     (nyi)	
                                     ;		
                                     0
                                     (nyi)	
                                     (nyi)	
                                     (nyi)	
                                     (nyi)	
                                     #f	
                                     (nyi)	
                                     (nyi)	
                                     (nyi)	
                                     #t))	

(define ρ (map cons native-names native-names))
(define σ (map cons native-names native-implementations))

(let loop ((state (rep "Small-Step Slip" ρ σ)))
  (loop (step state)))